#!/usr/bin/env python
"""
xmlrpccgi.py - A CGI XML-RPC server.
$Id: xmlrpccgi.py,v 1.7 2001/03/09 20:32:57 kalle Exp $

Loosely based on xmlrpcserver.py by Fredrik Lundh.

Copyright (c) 2001 Kalle Svensson (kalle@gnupung.net)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
"""

import xmlrpclib
import sys, os, string, types, time

class ProtocolError(Exception):
    def __init__(self, code, message):
        Exception.__init__(self, code, message)
        self.code = code
        self.message = message

class Server:
    def __init__(self, log = sys.stderr):
        self.log = log
        self.methods = {}
        self.register_method("system.listMethods", self.list_methods)
        self.register_method("system.methodDoc", self.method_doc)

    def register_method(self, name, func):
        """
        Register a method with the server.
        """
        parts = string.split(name, ".")
        d = self.methods
        for part in parts[:-1]:
            if d.has_key(part):
                if not type(d[part]) == types.DictType:
                    raise ValueError, "Key already exists and is non-dict"
            else:
                d[part] = {}
            d = d[part]
        d[parts[-1]] = func

    def get_method(self, name):
        """
        Return the method associated with a name.
        """
        parts = string.split(name, ".")
        d = self.methods
        for part in parts[:-1]:
            d = d[part]
        return d[parts[-1]]

    def call_method(self, method, params):
        """
        Call the method with params.
        """
        try:
            return apply(self.get_method(method), params)
        except KeyError:
            raise AttributeError, "No such method: %s" % method

    def list_methods(self, methods = None):
        """
        array listMethods()
        Get a list of the available methods.
        Automatically registered as system.listMethods()
        """
        if methods == None:
            methods = self.methods
        res = []
        for key, val in methods.items():
            if type(val) == types.DictType:
                for x in self.list_methods(val):
                    res.append("%s.%s" % (key, x))
            else:
                res.append(key)
        return res

    def method_doc(self, method):
        """
        string methodDoc(string)
        Returns the function documentation.
        """
        try:
            m = self.get_method(method)
            if m.__doc__:
                return m.__doc__
            else:
                return ""
        except KeyError:
            raise AttributeError, "No such method: %s" % method

    def get_request(self):
        """
        Get the XML-RPC request, or raise a ProtocolError on invalid input.
        """

        try:
            method = os.environ["REQUEST_METHOD"]
            type = os.environ["CONTENT_TYPE"]
            length = os.environ["CONTENT_LENGTH"]
        except KeyError:
            raise ProtocolError(400, "Bad Request")
        
        if string.lower(method) != "post":
            raise ProtocolError(405, "Method Not Allowed")
        if string.lower(type) != "text/xml":
            raise ProtocolError(400, "Bad Request")
        if int(length) <= 0:
            raise ProtocolError(411, "Length Required")

        data = sys.stdin.read(int(length))

        if len(data) != int(length):
            raise ProtocolError(400, "Bad Request")

        return data

    def get_response(self, data):
        """
        Get a XML-RPC response.
        """

        params, method = xmlrpclib.loads(data)
        try:
            response = self.call_method(method, params)
            if type(response) != types.TupleType:
                response = (response,)
        except:
            response = xmlrpclib.dumps(xmlrpclib.Fault(1,
                                                       "%s:%s" % (sys.exc_type,
                                                                  sys.exc_value)))
        else:
            response = xmlrpclib.dumps(response, methodresponse=1)
        return response

    def handle_error(self, code, message, log = 0, escape = 1):
        if log:
            exc_info = sys.exc_info()
        if type(code) != types.IntType:
            # If code is not an integer, we could perhaps fall victim
            # to cross-site scripting attacks.
            try:
                code = int(code)
            except:
                code = 500
                message = "Internal Server Error"
                self.log_string("*** Could not convert error code to int! ***")
        message = xmlrpclib.escape(message, 1)
        print """Status: %3d %s
Content-type: text/html

<html>
<head>
<title>%3d %s</title>
</head>
<body>
<h1>%3d %s</h1>
<p>Unexpected error processing XML-RPC request.</p>
</body>
</html>
""" % ((code, message) * 3)
        if log:
            self.log_exception(exc_info, escape)
            del exc_info

    def log_string(self, s, escape = 0):
        if "\n" in s:
            s = ("[%s] [xmlrpccgi]\n" %
                 time.asctime(time.localtime(time.time()))) + s
        else:
            s = ("[%s] [xmlrpccgi] " %
                 time.asctime(time.localtime(time.time()))) + s
        if escape:
            s = xmlrpclib.escape(s)
        self.log.write(s + "\n")

    def log_exception(self, exc_info = None, escape = 1):
        if not exc_info:
            exc_info = sys.exc_info()
        import traceback
        s = string.join(apply(traceback.format_exception, exc_info), "")
        self.log_string(s[:-1], escape) # chop off newline
        exc_info = None

    def run(self):
        """
        Get request, response and print result.
        """

        try:
            data = self.get_request()
            response = self.get_response(data)
        except ProtocolError, e:
            self.handle_error(e.code, e.message, 0)
        except:
            # internal error, report as HTTP server error
            self.handle_error(500, "Internal Server Error", 1)
        else:
            # got a valid XML-RPC response
            print """Status: 200 OK
Content-type: text/xml
Content-length: %d
""" % len(response)
            print response

### Test code below...

def arrayOfStructsTest(arr):
    """
    validator1.arrayOfStructsTest (array) returns number

    This handler takes a single parameter, an array of structs, each
    of which contains at least three elements named moe, larry and
    curly, all <i4>s.  Your handler must add all the struct elements
    named curly and return the result.
    """

    sum = 0
    for struct in arr:
        sum = sum + struct["curly"]
    return sum

def countTheEntities(s):
    """
    validator1.countTheEntities (string) returns struct

    This handler takes a single parameter, a string, that contains any
    number of predefined entities, namely <, >, &, ' and \".

    Your handler must return a struct that contains five fields, all
    numbers: ctLeftAngleBrackets, ctRightAngleBrackets, ctAmpersands,
    ctApostrophes, ctQuotes.

    To validate, the numbers must be correct.
    """

    trans = {"ctLeftAngleBrackets" : "<",
             "ctRightAngleBrackets" : ">",
             "ctAmpersands" : "&",
             "ctApostrophes" : "'",
             "ctQuotes" : '"'}
    counts = {"ctLeftAngleBrackets" : 0,
              "ctRightAngleBrackets" : 0,
              "ctAmpersands" : 0,
              "ctApostrophes" : 0,
              "ctQuotes" : 0}
    for count in counts.keys():
        for c in s:
            if c == trans[count]:
                counts[count] = counts[count] + 1
    return counts

def easyStructTest(struct):
    """
    validator1.easyStructTest (struct) returns number

    This handler takes a single parameter, a struct, containing at
    least three elements named moe, larry and curly, all <i4>s. Your
    handler must add the three numbers and return the result.
    """

    return struct["moe"] + struct["larry"] + struct["curly"]

def echoStructTest(struct):
    """
    validator1.echoStructTest (struct) returns struct

    This handler takes a single parameter, a struct. Your handler must
    return the struct.
    """

    return struct

def manyTypesTest(i, b, s, d, dt, b64):
    """
    validator1.manyTypesTest(number, boolean, string, double, dateTime, base64)
        returns array

    This handler takes six parameters, and returns an array containing
    all the parameters.
    """

    return [i, b, s, d, dt, b64]

def moderateSizeArrayCheck(arr):
    """
    validator1.moderateSizeArrayCheck (array) returns string

    This handler takes a single parameter, which is an array
    containing between 100 and 200 elements. Each of the items is a
    string, your handler must return a string containing the
    concatenated text of the first and last elements.
    """

    return arr[0] + arr[-1]

def nestedStructTest(struct):
    """
    validator1.nestedStructTest (struct) returns number

    This handler takes a single parameter, a struct, that models a
    daily calendar. At the top level, there is one struct for each
    year. Each year is broken down into months, and months into
    days. Most of the days are empty in the struct you receive, but
    the entry for April 1, 2000 contains a least three elements named
    moe, larry and curly, all <i4>s. Your handler must add the three
    numbers and return the result.

    Ken MacLeod: "This description isn't clear, I expected
    '2000.April.1' when in fact it's '2000.04.01'. Adding a note
    saying that month and day are two-digits with leading 0s, and
    January is 01 would help." Done.
    """

    elem = struct["2000"]["04"]["01"]
    return elem["moe"] + elem["larry"] + elem["curly"]

def simpleStructReturnTest(num):
    """
    validator1.simpleStructReturnTest (number) returns struct

    This handler takes one parameter, and returns a struct containing
    three elements, times10, times100 and times1000, the result of
    multiplying the number by 10, 100 and 1000.
    """

    return {"times10"   : num * 10,
            "times100"  : num * 100,
            "times1000" : num * 1000}

def register_test_methods(s):
    s.register_method("validator1.arrayOfStructsTest", arrayOfStructsTest)
    s.register_method("validator1.countTheEntities", countTheEntities)
    s.register_method("validator1.easyStructTest", easyStructTest)
    s.register_method("validator1.echoStructTest", echoStructTest)
    s.register_method("validator1.manyTypesTest", manyTypesTest)
    s.register_method("validator1.moderateSizeArrayCheck", moderateSizeArrayCheck)
    s.register_method("validator1.nestedStructTest", nestedStructTest)
    s.register_method("validator1.simpleStructReturnTest", simpleStructReturnTest)

def test():
    server = Server()
    register_test_methods(server)
    server.run()

if __name__ == "__main__":
    test()
